/*
  This is comands.c

  Coxeter version 1.0  Copyright (C) 1991 Fokko du Cloux
  See file main.c for full copyright notice
*/

#include <stdio.h>

#include "commands.h"

/*

      This is a command-driven version of the interface. This will allow for
  a much easier maintenance of the interface as the program grows. We have
  tried to keep the philosophy that the user should be protected from his
  worst mistakes. This means in particular that he is stopped right away
  when the choices he makes will cause overflow for the types that are
  involved. We have also tried to roughly assess the size of a required
  computation, so that the user may be warned when he's in for a few hours.
  In any case, except for possible bugs, the program should never crash.
  If it must, it should exit gracefully, usually because the available
  memory has been exhausted.

      There are several important modifications with respect to previous
  versions. A status data structure is maintained at all times, so that the
  program can keep track of what has been done, and avoids duplication of
  work. In this way, it is also possible to change rank and/or type without
  exiting the program. Also, it is possible to dump the result of the KL-
  computation in the form of binary files, which are the ones that the
  computer keeps in memory when the program is running. This allows the
  data to be accessed directy from the files, and therefore avoids having
  to recompute them each time. This is especially important for large
  groups. (Actually, whe have four sizes of groups : small, which includes
  A_n for n up to 6, B_n,C_n and D_n for n up to 5, F4, and G2; medium,
  including A7,B6,C6,D6,E6 and H4; large, including A8,A9,B7,C7,D7,E7; and
  huge, which is everything else. Previously, kl-computations were possible
  only for small groups; now, on a reasonable workstation, all the medium 
  ones are covered.)

      At the level of input, we have introduced a few shorthand notations.
  In particular, products and powers can be input directly, and elements
  can be input (and computed with) as weylnbrs. A few help-messages have
  been included. Output to files is still in TeX-format.

*/

Run()

/*
  This function runs the program. It is in this module, so that all
  command-related definitions can be kept in command.h, to improve
  maintenability.
*/

{
  
  int command;

  MakeCommandTree();
  MakeResponse();

  status.mode = general_m;
  prev_mode = general_m;
  prev_command = undef_c;

  status.type = '\000';
  status.rank = 0;

  status.size = 0;
  
  printf("This is Coxeter version %s.\n\
Enter help if you need assistance, carriage return to start the program.\n\n",
  VERSION);

  while (1)
    {
      Prompt();
      command = GetCommand();
      if (command != _c)
	{
	  prev_command = command;
	}
      response[command]();
    }
  
}    /* end of Run */


static GetCommand()

/*
  Gets the command from the user. Command completion is implemented, as long as it
  is unambiguous -- but the command must be an initial segment in the full name.
  This seems safer than to allow arbitrary extensions.
*/

{

  char buf[LINESIZE];
  int command;

  while (1)
    {
      gets(buf);
      command = FindCommand(comm_root,buf);
      switch (command)
	{
	case undef_c:
	  printf("\t%s : command not found\n",buf);
	  Prompt();
	  break;
	case ambig_c:
	  printf("\t%s : ambiguous\n",buf);
	  Prompt();
	  break;
	default:
	  return(command);
	}
    }
  
}    /* end of GetCommand */


static FindCommand(cc,b)

/*
  Finds the string b on the tree pointed by cc, or returns -1 if the string is not
  found.
*/

     ccell_t *cc;
     char *b;

{
  
  if (*b == 0)            /* command was found */
    return(cc->command);

  if ((cc = cc->left) == 0)  /* command was too long */
    {
      return(undef_c);
    }

  while ((cc->right)&&(*b>cc->letter))
    {
      cc = cc->right;
    }

  if (*b == cc->letter)
    {
      return(FindCommand(cc,b+1));
    }
  else
    {
      return(undef_c);
    }
  
}    /* end of GetCommand */


static CarriageReturn()

/*
  Response to a carriage return. As a rule, it will repeat the previous
  command.
*/

{

  switch (status.mode)
    {
    case general_m:
      if (status.type == 0)
        {
        prev_command = undef_c;
        Start();
        }
      break;
    case help_m:
      break;
    default:
      if (prev_command == undef_c)
	{
	  ;
	}
      else
	{
	  response[prev_command]();
	}
      break;
    }

}    /* end of CarriageReturn */


static Author()

/*
  Prints the author's adress.
*/

{

  switch (status.mode)
    {
    case help_m:
      printf("%s",author_mess);
      break;
    default:
      printf("%s",acknowledgement);
      break;
    }

}    /* end of Author */


static Cell()

/*
  Response to cell.
*/

{

  switch (status.mode)
    {
    case kl_m:
      GetCell();
      break;
    case help_m:
      printf("%s",cell_mess);
      break;
    default:
      printf("\tYou should say \"kl\" first\n");
      break;
    }

}    /* end of Cell */


static Compute()

/*
  Response to comp.
*/

{
  weylnbr_t w;

  switch (status.mode)
    {
    case help_m:
      printf("%s",comp_mess);
      break;
    case general_m:
      if (status.type == 0)
	{
	  printf("\tYou have to set the type and rank first\n\n");
	  Start();
	  printf("\n");
	}
      Weyl();
      status.weyldone = 1;
      status.mode = weyl_m;
    default:
      printf("compute : ");
      GetInput(input_array);
      PrintArray(input_array,stdout);
      ArrayToWeylnbr(input_array,&w);
      printf("  (#%ld)",w);
      printf("\n");
      break;
    }

}    /* end of Compute */


static HelpMe()

/*
  Response to help request.
*/

{
  int j;

  if (status.mode == help_m)
    {
      printf("%s",help_mess);
      return(0);
    }

  prev_mode = status.mode;    /* remember current mode */
  printf("%s",help1_mess);

  for (j=0;j<MAXCOMMAND;j++)
    {
      printf("%s",allcmds[j]);
    }

  printf("%s",help2_mess);

  status.mode = help_m;

}    /* end of HelpMe */


static Info()

/*
  Response to info.
*/

{

  printf("%s",info_mess);

}    /* end of Info */


static KLResponse()

/*
  Response to kl.
*/

{
  weylnbr_t m;
  int l;

  switch (status.mode)
    {
    case help_m:
      printf("%s",kl_mess);
      break;
    case kl_m:
      printf("\tYou are already in kl mode\n");
      break;
    case general_m:
      if (status.type == '\000')
	{
	  printf("\tYou have to set the type and rank first\n\n");
	  Start();
	}
    default:
      if (status.kldone == 0)
	{
	  /* check for overflow */
	  
	  m = 0;
	  m = ~m;  /* largest possible value */

	  while (l = CheckChoice(status.type, status.rank, m))
	    {
	      printf("\nRank too big -- %d is the limit\n",l-1);
	      NewRank();
	    }

      printf("\nComputing the Kazhdan-Lusztig file (this may take some time ...)\n");
	}

      status.mode = kl_m;

      if (status.size == small_size)
	{
	  dump_flag = 0;
	}
      else
	{
	  dump_flag = 1;
	}
      
      if (status.weyldone == 0)
	{
	  Weyl();
	  status.weyldone = 1;
	}
      if (status.bruhatdone == 0)
	{
	  PreBruhat();
	  status.bruhatdone = 1;
	}
      if (status.kldone == 0)
	{
	  KL();
	  status.kldone = 1;
	}
      printf("\n");
      break;
    }

  return(0);

}    /* end of KLResponse */


static Pol()

/*
  Response to pol.
*/
     
{

  switch (status.mode)
    {
    case help_m:
      printf("%s",pol_mess);
      break;
    case kl_m:
      GetPol();
      break;
    default:
      printf("\tYou should say \"kl\" first\n");
      break;
    }

}    /* end of Pol */


static Presentation()

/*
  Response to pres.
*/

{

  switch (status.mode)
    {
    case help_m:
      printf("%s",pres_mess);
      break;
    default:
      if (status.weyldone == 0)
	{
	  Weyl();
	  status.weyldone = 1;
	}
      CompletePresentation();
      printf("\tpresentation written in %c%d_presentation\n",
      status.type,status.rank);
      break;
    }

}    /* end of Presentation */


static QQuit()

/*
  Response to qq
*/

{

  exit(0);

}    /* end of QQuit */


static Quit()

/*
  Response to quit
*/

{

  switch (status.mode)
    {
    case general_m:
      exit(0);
    case help_m:
      status.mode = prev_mode;  /* revert to previous mode */
      prev_command = undef_c;
      break;
    default:
      status.mode = general_m;
      break;
    }

}    /* end of Quit */


static NewRank()

/*
  Response to rank.
*/

{
  weylsize_t m;
  int l;

  switch (status.mode)
    {
    case help_m:
      printf("%s",rank_mess);
      break;
    default:
      status.mode = general_m;  /* start from scratch */

      status.weyldone = 0;
      status.bruhatdone = 0;
      status.kldone = 0;

      if (status.type == 0)
	{
	  printf("\tYou have to set the type first\n");
	  NewType();
	  return(0);
	}
  
      GetRank(&rank);
      status.rank = rank;
      MakeSize();

      /* check for overflow */

      m = 0;
      m = ~m;  /* maximal unsigned quantity of its size */
      
      while (l = CheckChoice(status.type,status.rank,m))
	{
	  printf("\nRank too big -- %d is the limit\n",l-1);
	  GetRank(&rank);
	  status.rank = rank;
	}
      
      InitCommands();
      InitInterface();
      break;
    }

}    /* end of NewRank */


static NewRepresentation()

/*
  Response to rep.
*/

{

  switch(status.mode)
    {
    case help_m:
      printf("%s",rep_mess);
      break;
    case general_m:
      if (status.type == 0)
	{
	  printf("\tYou have to set the type and the rank first");
	  NewType();
	}
    default:
      GetRepresentation(&representation);
      status.representation = representation;
      break;
    }

}    /* end of NewRepresentation */


static Start()

/*
  Response to run.
*/

{
  
  switch (status.mode)
    {
    case help_m:
      printf("%s",rep_mess);
      break;
    default:
      status.mode = general_m;  /* start from scratch */

      printf("Enter the type and rank you're interested in :\n");

      NewType();
      GetRepresentation(&representation);

      status.representation = representation;
      break;
    }
  
}				/* end of Start */


static Tex()

/*
  Response to tex.
*/

{

  switch (status.mode)
    {
    case help_m:
      printf("%s",tex_mess);
      break;
    case kl_m:
      GetAll();
      printf("\n");
      break;
    default:
      printf("\tYou should say \"kl\" first\n");
      break;
    }

}    /* end of Tex */


static NewType()

/*
  Response to type.
*/

{

  switch (status.mode)
    {
    case help_m:
      printf("%s",type_mess);
      break;
    default:
      GetType(&type);
      status.type = type;
      NewRank();
      break;
    }
  
}				/* end of NewType */


static VermaResponse()

/*
  Response to Verma
*/

{

  switch(status.mode)
    {
    case help_m:
      printf("%s",verma_mess);
      break;
    case kl_m:
      GetMultiplicities();
      break;
    default:
      printf("\tYou should say \"kl\" first\n");
      break;
    }

}    /* end of VermaResponse */


static Prompt()

/*
  Prints a prompt.
*/

{

  switch (status.mode)
    {
    case general_m:
      printf("coxeter : ");
      break;
    case weyl_m:
      printf("weyl : ");
      break;
    case bruhat_m:
      printf("bruhat : ");
      break;
    case kl_m:
      printf("kl : ");
      break;
    case help_m:
      printf("help : ");
      break;
    }
  
}    /* end of prompt */


static MakeSize()

/*
  Fills in status.size, after type and rank have been set.
*/

{

  switch (status.type)
    {
    case 'A':
      if (status.rank <= 6)
	{
	  status.size = small_size;
	}
      else if (status.rank <= 7)
	{
	  status.size = medium_size;
	}
      else if (status.rank <= 9)
	{
	  status.size = large_size;
	}
      else
	{
	  status.size = huge_size;
	}
      break;
    case 'B':
    case 'C':
      if (status.rank <= 5)
	{
	  status.size = small_size;
	}
      else if (status.rank <= 6)
	{
	  status.size = medium_size;
	}
      else if (status.rank <= 8)
	{
	  status.size = large_size;
	}
      else
	{
	  status.size = huge_size;
	}
      break;
    case 'D':
      if (status.rank <= 5)
	{
	  status.size = small_size;
	}
      else if (status.rank <= 6)
	{
	  status.size = medium_size;
	}
      else if (status.rank <= 8)
	{
	  status.size = large_size;
	}
      else
	{
	  status.size = huge_size;
	}
      break;
    case 'E':
      if (status.rank <= 5)
	{
	  status.size = small_size;
	}
      else if (status.rank = 6)
	{
	  status.size = medium_size;
	}
      else if (status.rank = 7)
	{
	  status.size = large_size;
	}
      else
	{
	  status.size = huge_size;
	}
      break;
    case 'F':
      status.size = small_size;
      break;
    case 'G':
      status.size = small_size;
      break;
    case 'H':
      if (status.rank <= 3)
	{
	  status.size = small_size;
	}
      else
	{
	  status.size = medium_size;
	}
      break;
    default:  /* should never be reached */
      printf("\nMakeSize : invalid type\n");
      exit(0);
      break;
    }

}    /* end of MakeSize */


static MakeCommandTree()

/*
  Constructs the tree on which the commands are found. The tree is organized as
  a binary tree. The entry point, comm_root, corresponds to the "empty" command
  (which is actually used very often, with different meanings in different
  contexts). The left son of a given cell is the first son (ie the first extension
  of the string into a longer string); the right son is the first brother.

  The tree is constructed from the data in comm_tdata, comm_nbr and comm_letter.
  Those are three arrays, with an entry for each cell in the tree. The first one
  describes the combinatorial structure of the tree, by giving for each cell
  the number of its first-generation descendants (it can be checked that this
  entirely determines the tree). The second and third give the command- and letter-
  field respectively, for each cell.
*/

{

  ccell_t *cc,*firstfree;
  int k,j;
  
  CALLOC(comm_root,MAXCELL,ccell_t,;)
  
  cc = comm_root;
  firstfree = comm_root+1;

  for (k=0;k<MAXCELL;k++,cc++)
    {
      cc->command = comm_nbr[k];
      cc->letter = comm_letter[k];
      if (comm_tdata[k])	/* cc has descendants */
	{
	  cc->left = firstfree;
	  for (j=1;j<comm_tdata[k];j++)
	    {
	      firstfree->right = firstfree+1;
	      firstfree++;
	    }
	  firstfree++;
	}
    }
  
}    /* end of MakeCommandTree */


static InitCommands()

/*
  Initializations for the command module. Must be called each time rank changes.
*/

{

  CALLOC(input_array,rank,weylpiece_t,;)
  
}    /* end of InitCommands */


static MakeResponse()

/*
  Sets the responses to the various commands 
*/

{

  response[_c] = CarriageReturn;
  response[author_c] = Author;
  response[cell_c] = Cell;
  response[comp_c] = Compute;
  response[help_c] = HelpMe;
  response[info_c] = Info;
  response[kl_c] = KLResponse;
  response[pol_c] = Pol;
  response[pres_c] = Presentation;
  response[qq_c] = QQuit;
  response[quit_c] = Quit;
  response[rank_c] = NewRank;
  response[rep_c] = NewRepresentation;
  response[tex_c] = Tex;
  response[type_c] = NewType;
  response[verma_c] = VermaResponse;

}    /* end of MakeResponse */
