/*
  This is kl.c

  Coxeter version 1.0  Copyright (C) 1991 Fokko du Cloux
  See file main.c for full copyright notice
*/

#include <stdio.h>
#include "kl.h"

/*
  This module constructs the Kazhdan-Lusztig polynomials.
  
  The main list is called kl_list. Each row in this list,
  corresponding to w, say, has a cell of type kldata_t for each
  y <= w in Bruhat order which is "extremal" w.r.t. w, i.e. which
  is taken down by each generator s which takes w down on left or 
  right. The work done by Bruhat() is mainly to construct this list
  efficiently; in principle everything could be done (slowly) from
  the extremal-list and the Weyl group routines.
  
  The reason for restricting to extremal pairs is that if for
  example l(sw)<l(w), l(sy)>l(y), then P_{y,w}=P_{sy,w} (see the
  paper by Kazhdan and Lusztig). So all the other kl-polynomials
  can be easily extracted from this list, if needed; this is what 
  KLpol() does. Moreover, if \mu_{y,w}\neq 0 and l(w)-l(y)>1, it
  is clear that (y,w) must be an extremal pair (the cases for which
  l(w)-l(y)=1 are exactly those in prbruhat[w]). So if one is 
  interested in left and right cells one should also concentrate on 
  extremal pairs (this is not implemented yet).
  
  The other idea is to refer to the kl-polynomials by address.
  The point is that there will be a lot of repetitions among them; so
  an index is kept of all the polynomials met so far, and each is 
  written down only once. Although some relative adressing scheme is 
  certainly possible and would allow to cut the size of klpol_t at least 
  in half, for simplicity we have chosen not to do so. This might change 
  in the future when considering larger groups.
  
  Otherwise the algorithm is simply taken from the Inventiones paper 
  of Kazhdan and Lusztig. We have tried to use the full recursion formula 
  only when there is no other way, and not to repeat the same calculation 
  too many times.
*/

KL()
     
     /********************************************************************
       
       Constructs kl_list, performing all the necessary initializations
       
       *********************************************************************/
     
{
  InitConstants();

  if (ReadFiles())  /* files not available */
    {

      /* dumping should be made optional in some later version */
      dump_flag = 1;

      InitKL();
      MakeDFLists(dflist);
      MakeExtrList();
      MakeKLList();
      MakeBCells();
      if (dump_flag)
	{
	  DumpFiles();
	  ReadFiles();
	}
    }
  else
    {
      MakeBCells();
    }
}				/* end of KL */


PreKL()

/*
  Regroups the preliminaries to KL, for separate use
*/

{

  InitConstants();
  InitKL();
  MakeDFLists(dflist);

}    /* end of PreKL */


static InitConstants()

/*

  Initializes the following constants :
  
     - wrowlength : length (in lflags_t) of a bit-array of length wsize;

     - betti_work : buffer for betti-polynomials; it has room for
       maxlength+1 coefficients, plus one for the degree;

     - bn_work : same thing.

*/

{
  mudata_t *mu;
  unsigned long m;

  CALLOC(kl_list,wsize,rowhead_t,;)    /* a message should go here */

  wrowlength = wsize/wordsize + 1;

  CALLOC(betti_work,maxlength+2,weylsize_t,;)
  CALLOC(bn_work,maxlength+2,weylsize_t,;)
  CALLOC(bklmult_work,maxlength+2,weylsize_t,;)

  CALLOC(mu,wsize,mudata_t,;)
  mubuffer.row = (char *)mu;

  m=1;
  m <<= (2*rank-1);

  CALLOC(bettipol,m,bettipol_t,;)

  bettizone = AllocZone(sizeof(weylsize_t));

  CALLOC(bettistore,1,betticell_t,;)
  bettistore->pol = (weylsize_t *)bettizone->firstfree;
  bettistore->pol[1] = 1;
  bettizone->firstfree += 2*sizeof(weylsize_t);
  bettizone->freestore -= 2;
}


InitKL()
     
/*
  Allocates memory for the following pointers :
  
    - kl_list, which is a pointer to wsize rowhead_t; note that the
      individual rows are NOT allocated at this point; this will be
      done by MakeExtrList;

    - dflist, which points to 2*rank rows of wsize bits;
      
    - klstore, which is a binary tree on which the different kl-
      polynomials are stored; the actual polynomials are written
      in klzone (this allows for variable size).

  It also creates the first cell in klstore, corresponding to the
  polynomial 1.
*/
     
{
  rankid_t l;
  
  CALLOC(dflist,2*rank,lflags_t *,;)

  for (l=0;l<2*rank;l++)
    {
      CALLOC(dflist[l],wrowlength,lflags_t,;)
    }
  
  klzone = AllocZone(sizeof(klcoeff_t));
  
  /* creation of first cell : pol = 1*/
  
  CALLOC(klstore,1,klcell_t,;)
  klstore->pol = (klcoeff_t *)klzone->firstfree;
  klstore->pol[1] = 1;
  klzone->firstfree += 2*sizeof(klcoeff_t);
  klzone->freestore -= 2;
  
}	/* end of InitKL */


MakeExtrList()

/*
  In this new version of MakeExtrList, we do not assume that BrOrder has
  already been constructed. Instead, we use a "floating" Bruhat order, i.e.
  we construct it and forget it as we go -- of course this is aimed
  towards groups for which the Bruhat-order is too large to keep in
  memory. Later on we will introduce a "status" structure that will
  enable us to be more flexible about this, allowing us to use faster
  versions of the KL-construction if there is enough memory available.

  For now, we will have to do without the bruhat-order after here; the
  part of it that is needed for the KL-construction will have to be
  recovered from the list of extremal pairs (i.e., to see if y <= w,
  make y extremal w.r.t. w, and see if it appears in kl_list[w]). This
  will of course be awfully slow.

  Another innovation is that we compute the number of elements in each row
  row of kl_list before creating the row. This avoids having to use the
  complicated memory management scheme.
*/

{
  lengthcount_t p;
  lflags_t **brpiece,**nextpiece,f;
  weylsize_t k,b,bp;
  lrowsize_t elsize,eltot;
  rowsize_t elrow,r,j;
  rowhead_t *klrow;
  kldata_t *kl;
  weylnbr_t w,y;

  CALLOC(brpiece,betti[0],lflags_t *,;)
  CALLOC(*brpiece,betti[0]*rowlength[0],lflags_t,;)
  **brpiece = wordmask[0];

  b = 0;    /* origin of the current piece */
  w = 0;    /* current row */
  eltot = 0;
 
  for (p=0;p<maxlength;p++)
    {
      CALLOC(nextpiece,betti[p+1],lflags_t *,;)
      CALLOC(*nextpiece,betti[p+1]*rowlength[p+1],lflags_t,;)
	/* a message should go here */

      MakeBrPiece(p+1,brpiece,nextpiece);
      Extremals(p,brpiece);

      elsize = 0;    /* size of the current piece of the KLlist */
      r = rowlength[p];
      bp = betti[p];
      klrow = kl_list + b;

      for (k=0;k<bp;k++,klrow++,w++)
	{
	  if (w > inverse[w])
	    {
	      elrow = 0;
	    }
	  else
	    {
	      elrow = RowBits(brpiece[k],r);
	    }
	  klrow->size = elrow;
	  elsize += elrow;
	}


      klrow = kl_list + b;
      CALLOC(kl,elsize,kldata_t,printf("\n\tp = %d",p);)

      for (k=0;k<bp;k++,klrow++)
	{
	  if (klrow->size == 0)
	    continue;
	  
	  (klrow)->row = (char *)kl;

	  for (j=0;j<r;j++)
	    {
	      f = brpiece[k][j];
	      y = j*BITS(lflags_t);
	      
	      for(;f!=0;f>>=1,y++)
		{
		  if (f%2)
		    {
		      (kl++)->y = y;
		    }
		}
	    }
	}

      FREE(*brpiece,betti[p]*rowlength[p],lflags_t)
      FREE(brpiece,betti[p],lflags_t *)
      brpiece = nextpiece;
      b += bp;
      eltot += elsize;
    }

  /* last row */
  
  klrow->size = 1;
  CALLOC(kl,1,kldata_t,;)
  klrow->row = (char *)kl;
  kl->y = w;

  eltot++;

}    /* end of MakeExtrList */


Extremals(p,br)

/*
  Given p and br (pointing to the p-th part of the bruhat-order), this
  function returns in br a bit-map of the extremal pairs. It is assumed
  that p points to betti[p] rows of length rowlength[p].
*/

     lengthcount_t p;
     lflags_t ** br;

{
  rowsize_t r,j;
  downflags_t df,df1;
  weylnbr_t k,bp,b;
  lengthcount_t q;
  rankid_t l;
  lflags_t *lf;

  r = rowlength[p];
  bp = betti[p];

  b = 0;
  for (q=0;q<p;q++)
    {
      b += betti[q];
    }

  for (k=0;k<bp;k++)
    {
      df = downflags[b+k];
      lf = br[k];
      
      for (j=0;j<r;j++,lf++)
	{
	  df1 = df;

	  for (l=0;l<rank;l++)    /* left downflags */
	    {
	      if (df1%2)
		{
		  *lf &= dflist[l+rank][j];
		}
	      df1 >>=1;
	    }

	  for (l=0;l<rank;l++)    /* right downflags */
	    {
	      if (df1%2)
		{
		  *lf &= dflist[l][j];
		}
	      df1 >>= 1;
	    }
	}
    }
}    /* end of Extremals */


rowsize_t RowBits(lf,r)

/*
  Given a pointer lf to a row of r lflags_t, this function returns the
  total number of set bits.
*/

     lflags_t *lf;
     rowsize_t r;

{
  rowsize_t count,j;

  count = 0;

  for(j=0;j<r;j++)
    {
      count += BitCount(*(lf++));
    }
  
  return(count);
}				/* end of RowBits */


MakeDFLists(dflist)

/*

  This function constructs 2*rank bit-arrays of length wsize; bits in array
  s (resp. rank+s) are set for those elements for which s*w < w (resp> w*s < w).
  It assumes dflist has been allocated and points to an array of 2*rank such
  bit-arrays.

*/

     lflags_t **dflist;

{

rankid_t l;
downflags_t f;
weylsize_t w,j;
int j1;

j = 0;
j1 = 0;

for (w=0;w<wsize;w++)
  {
    f = downflags[w];

    for (l=rank;l<2*rank;l++)
      {
	if (f%2)
	  {
	    dflist[l][j] |= wordmask[j1];
	  }
	f >>= 1;
      }

    for (l=0;l<rank;l++)
      {
	if (f%2)
	  {
	    dflist[l][j] |= wordmask[j1];
	  }
	f >>= 1;
      }

    if (++j1 == wordsize)
      {
	j++;
	j1 = 0;
      }
  }

}    /* end of MakeDFLists */


MakeMuList(w,s)
     
/*
  Makes the list of mu-coefficients from the corresponding row in kl_list,
  for which moreover ys < y; these are exactly the terms needed in the
  recursion formula. The length of the list is written into mubuffer.size.
*/
     
     weylnbr_t w;
     generator_t s;
     
{
  kldata_t *kl;
  mudata_t *mu;
  weylnbr_t *pr;
  int d;
  lengthcount_t p,q;
  register klpol_t a;
  rowsize_t rowsize,rj,c;
  sflags_t fs;
  int iflag;

  iflag = 0;

  if (w > inverse[w])
    {
      w = inverse[w];
      iflag = 1;
    }
  
  mu = (mudata_t *)mubuffer.row;

  if (iflag)  /* flag s in left downflags */
    {
      fs = 1 << rank + s-1;
    }
  else  /* flag s in right downflags */
    {
      fs = 1 << s-1;
    }
  
  for (p=0;firstlength[p+1]<=w;p++)
    ;				/* put length of w in p */
  
  kl = (kldata_t *)kl_list[w].row;
  rowsize = kl_list[w].size;
  c = 0;
  
  for (rj=0,q=0;rj<rowsize;rj++,kl++)
    {
      while (firstlength[q+1]<=kl->y)
	q++;			/* put length of kl->y in q */
      if ((p-q) < 3)		/* trivial cases to be considered later */
	{
	  break;
	}
      a = *kl->pol;
      d = *a;
      if ((2*d == p-q-1)&&(downflags[kl->y]&fs)) /* make a new mucoefficient */
	{
	  if (iflag)
	    {
	      mu->y = inverse[kl->y];
	    }
	  else
	    {
	      mu->y = kl->y;
	    }
	  mu->height = d;
	  mu->mu = a[mu->height+1]; /* highest coefficient */
	  mu++;
	  c++;
	}
    }
  
  pr=(weylnbr_t *)(prbruhat[w].row);
  rowsize = prbruhat[w].size;
  
  for (rj=0;rj<rowsize;pr++,rj++) /* the trivial cases */
    if (downflags[*pr]&fs)
      {
	if (iflag)
	  {
	    mu->y = inverse[*pr];
	  }
	else
	  {
	    mu->y = *pr;
	  }
	mu->height = 0;
	mu->mu = 1;
	mu++;
	c++;
      }

  mubuffer.size = c;
  
}   /* end of MakeMuList */


MakeKLList()
     
     /**********************************************************************
       
       Fills in the "pol"-part in kl_list; it follows closely the original
       KL algorithm.
       
       ***********************************************************************/
     
{
  weylnbr_t w,y,w1,y1,w2,y2,*ws,*sw;
  register weylsize_t j;
  rowhead_t *klrow;
  rowsize_t rmu,rmumax,rj,rjmax;
  kldata_t *kl;
  mudata_t *mu;
  klcoeff_t mucoeff;
  int d,i;
  generator_t s,is;
  lengthcount_t p,q;
  struct CutData cdata;
  klpol_t pol,pol1,*polhandle;
  klcoeff_t *a,*a1;
  downflags_t f,fw;
  
  CALLOC(pol,maxlength+2,klcoeff_t,;)
  CALLOC(ws,rank,weylnbr_t,;)
  CALLOC(sw,rank,weylnbr_t,;)
	
  kl = (kldata_t *)kl_list->row;
  kl->pol = &(klstore->pol);
  
  for (p=1,w=1,klrow=kl_list+1;p<=maxlength;p++)
    {
      /* go thru kl_list */

      if (status.size > small_size)
	{
	  printf("\n\nlength = %d\n",p);
	}

      for (j=0;j<betti[p];j++,w++,klrow++)
	{
	  
	  if (status.size > small_size)
	    {
	      printf("[%d] ",w);
	    }

	  if (w > inverse[w])	/* does not need to be considered */
	    {
	      continue;
	    }
	  cdata = CutOff(w);
	  w2 = cdata.i;
		  
	  fw = downflags[w];
	  
	  ws = lexprod+2*rank*w;
	  sw = lexprod+2*rank*w+rank;
	  
	  MakeMuList(w2,cdata.s);
	  
	  kl=(kldata_t *)klrow->row;
          rjmax = klrow->size;
	  
	  for (q=0,rj=0;rj<rjmax;rj++,kl++)
	    {
	      
	      y=kl->y;
	      while (y>=firstlength[q+1])
		q++;		/* advance q to length of y */
	      if ((p-q)<3)	/* P_{y,w} = 1 for all the y's left*/
		{
		  for (;rj<klrow->size;rj++,kl++)
		    {
		      kl->pol = &(klstore->pol);
		    }
		  break;
		}
	      
	      {
		register j;
		for (j=0,a=pol;j<=maxlength+1;j++)
		  {
		    *(a++) = 0;
		  }
	      }
	      
	      for (is=0,f=fw;is<rank;f>>=1,is++)

 /* If there exists an s such that ws < w, but y is not <= ws, then we
    have P_{y,w} = P_{ys,ws} -- the other terms in the recursion formula
    drop out */

		{
		  if (f%2 == 0)	/* ws !< w */
		    continue;
		  w1 = ws[is];
		  if (KLpol(y,w1) == NULL) /* y !<= ws */
		    {
		      y1 = lexprod[2*rank*y+is];
		      kl->pol = KLpol(y1,w1);
		      goto nexty;
		    }
		}
	      
	      for (is=0;is<rank;f>>=1,is++)

 /* If there exists an s such that sw < w, but y is not <= sw, then we
    have P_{y,w} = P_{sy,sw} -- the other terms in the recursion formula
    drop out */
		
		{
		  if (f%2 == 0)	/* sw !< w */
		    continue;
		  w1 = sw[is];
		  if (KLpol(y,w1) == NULL) /* y !<= w1 */
		    {
		      y1 = lexprod[2*rank*y+rank+is];
		      kl->pol = KLpol(y1,w1);
		      goto nexty;
		    }
		}
	      
	      
	      /* apply recursion formula */
	      
	      LexRProd(y,s=cdata.s,&y2);
	      
	      pol1 = *KLpol(y2,w2);
	      a = pol+1;
	      a1 = pol1+1;
	      for (i=0;i<= *pol1;i++)
		{
		  *(a++) = *(a1++);
		}
	      
	      pol1 = *KLpol(y,w2);
	      a = pol+2;
	      a1 = pol1+1;
	      for (i=0;i<= *pol1;i++)
		{
		  *(a++) += *(a1++);
		}
	      
	      /* substract correcting terms */

	      mu = (mudata_t *)mubuffer.row;
	      rmumax = mubuffer.size;
	      
	      for (rmu=0;rmu<rmumax;rmu++,mu++)
		{
		  if (polhandle = KLpol(y,mu->y))	
		    /* there is a correcting term */
		    {
		      pol1 = *polhandle;
		      a = pol+mu->height+2;
		      a1 = pol1+1;
		      mucoeff = mu->mu;
		      for (i=0;i<= *pol1;i++)
			{
			  *(a++) -= mucoeff * *(a1++);
			}
		    }
		}
	      /* get degree of pol */
	      
	      for (d=(p-q-1)/2;pol[d+1]==0;d--)
		;
	      *pol = d;
	      kl->pol = GetAdress(pol);

    /* to be deleted */
	      if (kl->pol==0)  /* polynomial had negative coefficients */
		{
		  exit();
		}
    /* end */
	      
	      nexty :
	      continue;
	      
	    }
 	}
    }
  
  
}	/* end of MakeKLList */


klpol_t *KLpol(y,w)
     
/*
  Returns, for y and w in the Weyl group (written in lexorder), the
  corresponding Kazhdan-Lusztig polynomial. Actually, the "pol" term
  in kldata_t is a pointer to a Kazhdan-Lusztig polynomial; this will
  be useful when we will have to convert the absolute adresses into
  relative ones when writing the results to a file.

  We assume kl_list has been constructed; in fact, this is essentially a 
  function for accessing kl_list. The polynomial is written as a list of 
  klcoeff_t, the first one equal to the degree (so the space occupied by a 
  klpol_t p is *p+2).
*/
     
     weylnbr_t y,w;
     
{
  
  kldata_t *kl0,*kl1,*kl;
  rowsize_t r;

  if (w > inverse[w])    /* use the fact that P_{y,w} = P_{y^{-1},w^{-1}} */
    {
      w = inverse[w];
      y = inverse[y];
    }

  if(w < y)
    {
      return(NULL);
    }
  
  y = Extr(y,w);		/* make y extremal w.r.t. w */

  if (w < y)
    {
      return(NULL);
    }
  
  kl0 = (kldata_t *)kl_list[w].row;
  kl1 = kl0 + kl_list[w].size;

  if (kl0->y == y)
    {
      return(kl0->pol);
    }
  
  for (r=kl_list[w].size;r>1;r=kl1-kl0)	    /* find y in kl_list */
    {
      kl = kl0 + r/2;
      if (kl->y == y)    /* y was found */
	{
	  return(kl->pol);
	}
      if (kl->y < y)
	{
	  kl0 = kl;
	}
      else
	{
	  kl1 = kl;
	}
    }
  
  return(NULL);    /* y was not found */
  
}				/* end of KLpol */


weylnbr_t Extr(y,w)
     
     /**********************************************************************
       
       For y <= w in the Weyl group, returns the unique extremal element
       w.r.t. w in the class of y.       
       ***********************************************************************/
     
     weylnbr_t y,w;
     
{
  register int j;
  register downflags_t f;
  
  if (w==0)
    {
      return(0);
    }
  
  while (f=downflags[w]&~downflags[y])
    {
      for (j=0;f%2==0;f>>=1)
	j++;
      y = lexprod[2*rank*y+j];
    }
  
  while (f);
  
  return(y);
  
}	/* end of Extr */


klpol_t *GetAdress(pol)
     
/*
  Given a klpol_t pol, returns its address in klstore, as a handle. If pol 
  is not found, a new cell is created in klstore, and its adress returned.
  klstore is organized as a binary tree; each cell contains the absolute
  adress of the corresponding polynomial, and the absolute adresses of
  the left and right sons.
       
*/
     
     klpol_t pol;
     
{
  register klcell_t *c;
  int m;
  
  c = klstore;	/* entry point of tree */
  
  while (m = Compare(pol,c->pol))
    {
      if (m > 0) /* pol is smaller than c->pol */
	{
	  if (c->left)	/* go to left son */
	    {
	      c = c->left;
	    }
	  else	/* make new cell */
	    {
	      c->left = MakeSon(pol);
	      c = c->left;
	      break;
	    }
	}
      else	/* pol is greater than c->pol */
	{
	  if (c->right)	/* go to right son */
	    {
	      c = c->right;
	    }
	  else	/* make new cell */
	    {
	      c->right = MakeSon(pol);
	      c = c->right;
	      break;
	    }
	}
    }
  
  /* if we get here, pol has been found */

  /* to be deleted */
  if (c==0)  /* polynomial had negative coefficients */
    {
      return(0);
    }
  /* end */

  return(&(c->pol));
  
}	/* end of GetAdress */


Compare(p,q)
     
     /********************************************************************
       
       Given two klpol_t's p and q, compare them in lexicographical order,
       first by degree, then starting from the low order term.
       
       *********************************************************************/
     
     klpol_t p,q;
     
{
  int i,m,d;
  
  if (m = *q - *p)	/* degrees are different */
    {
      return(m);
    }
  
  d = *p;
  for (i=0;i<=d;i++)
    if (m = *(++q)-*(++p))
      {
	return(m);
      }
  
  return(0);
  
}	/* end of Compare */


klcell_t *MakeSon(p)
     
     /********************************************************************
       
       Makes a leaf for klstore, and adds the polynomial p to the
       storezone, writing its address in the new cell.
       
       *********************************************************************/
     
     klpol_t p;
     
{
  klcell_t *newcell;
  int i;
  
  CALLOC(newcell,1,klcell_t,;)

  if (klzone->freestore <= *p+1) /* make room */
    {
      NewBlock(klzone,sizeof(klcoeff_t));
    }
  
  newcell->pol = (klpol_t)klzone->firstfree;
  klzone->firstfree += (*p+2)*sizeof(klcoeff_t);
  klzone->freestore -= *p+2;
  for (i=0;i<= *p+1;i++)
    (newcell->pol)[i] = p[i];
  
  return(newcell);
  
}				/* end of MakeSon */


BettiNumbers(w,h,ih)
     
/*
  This function constructs both the betti numbers and the IH betti
  numbers for the bruhat-cell of w. It uses kl_list directly, taking
  advantage of the fact that the kl-polynomial is the same for all
  the elements in an orbit of the subgroup of W\times W generated
  by the left and right downflags for w.

  In particular, both the betti and the IH-betti polynomials are
  divisible by the betti-polynomial of the subgroup generated by,
  say, left downflags. To compute the quotient, we have to determine
  for each y which right downflags take y to its orbit under left
  downflags; this is done in the function RtoL (later on, it may
  be worthwile to tabulate them). The betti-polynomial of the
  homogeneous space we're interested in is then found by a division.

  For the IH-homology case, we should beware of the fact that betti-
  polynomials and kl-polynomials do not necessarily have the same type.
*/
     
     weylnbr_t w;
     weylsize_t *h,*ih;
     
{
  kldata_t *kl;
  rowsize_t r,rmax;
  sflags_t lf,rf,rtolf;
  lengthcount_t p,q;
  int j,m;

  for (p=0;w>=firstlength[p+1];p++)  /* p = l(w) */
    ;

  for (q=0;q<=p;q++)  /* clean up h */
    {
      h[q] = 0;
    }
  for (q=0;q<=p;q++)  /* and ih */
    {
      ih[q] = 0;
    }

  kl = (kldata_t *)kl_list[w].row;
  rmax = kl_list[w].size;

  lf = downflags[w] >> rank;	/* lf flags the left downflags */
  rf = downflags[w] & ~(lf << rank); /* rf flags the right downflags */

  q = 0;

  for (r=0;r<rmax;r++,kl++)
    {
      for (j=0;j<=maxlength+1;j++)
	bn_work[j] = 0;
      while (kl->y >= firstlength[q+1]) /* advance q to length of y */
	q++;
      rtolf = RtoL(kl->y,lf,rf);
      BPolDiv(bettipol[rf],bettipol[rtolf],bn_work);

      /* need to shift by minimum length in cell */

      m = q - bettipol[lf][0]- bn_work[0];
      for (j=bn_work[0];j>=0;j--)
	{
	  bn_work[j+m+1] = bn_work[j+1];
	}
      for (j=0;j<m;)
	{
	  bn_work[++j] = 0;
	}
      *bn_work += m;

      BPolAdd(bn_work,h,h);
      
      BKLPolMult(bn_work,*kl->pol,bn_work);
      BPolAdd(bn_work,ih,ih);
    }
      
  BPolMult(bettipol[lf],h,h);
  BPolMult(bettipol[lf],ih,ih);

}				/* end of BettiNumbers */


sflags_t RtoL(y,lf,rf)

/*
  Returns the subset of generators s in rf for which there is a t in lf
  such that ys = ty. It is assumed that lf (resp. rf) is included in the
  left (resp. right) downflags for y.
*/

{
  register generator_t s,t;
  register sflags_t f;
  sflags_t rtolf;
  register weylnbr_t ys;

  rtolf = 0;

  for (s=0;rf;rf>>=1,s++)
    {
      if (1&~rf)
	continue;
      ys = lexprod[2*y*rank+s];
      for (f=lf,t=0;f;f>>=1,t++)
	{
	  if (1&~f)
	    continue;
	  if (lexprod[2*rank*ys+rank+t]==y)  /* tys = y */
	    {
	      rtolf += 1 << s;
	      break;
	    }
	}
    }

  return(rtolf);
}    /* end of RtoL */


MakeBCells()

/*
  This function constructs the "betti-cells". Recall that on each Bruhat-cell
  we have an action of a Coxeter-group, namely the subgroup of W \times W
  generated by the generators in downflags{w}. In kl_list, we find an entry
  for each orbit ot this action. Therefore, if we want to find, say, the
  betti-numbers of the Bruhat-cell, it is enough to add up the betti-numbers
  of each of the orbits, suitably shifted. Notice also that the action of the
  subgroup corresponding to, say, the left downflags, is free. Therefore the
  orbit is the product of this group and a homogeneous space for the right
  downflags. So the information we need is the betti-polynomial for each
  possible homogeneous space; these spaces are described by a pair of subsets
  T_1,T_2 of S = {1, ...,n} such that T_1 \subset T_2. This function makes a 
  table of the corresponding polynomials.
*/

{
  sflags_t f,maxf,fs,cs;
  generator_t s;
  typeid_t x;
  rankid_t l;

  maxf = 1 << rank-1;
  maxf += maxf-1;
  
  bettipol[0] = bettistore->pol;  /* bettipol[0] = 1 */

  for (f=1;f<=maxf;f++)
    {
      for (s=0,fs=f;fs;s++,fs>>=1)
	;
      cs = Component(f,s);
      if (cs!=f)
	{
	  BPolMult(bettipol[cs],bettipol[f&~cs],betti_work);
	  bettipol[f] = GetBAdress(betti_work);
	  continue;
	}

      /* now f is connected */
      
      fs = 1 << s-1;
      x = CompType(f);
      l = BitCount(f);
      *betti_work = GetMaxql(x,l);
      BettiData(x,l,betti_work+1);
      BPolMult(betti_work,bettipol[f&~fs],betti_work);
      bettipol[f] = GetBAdress(betti_work);
    }
}    /* end of MakeBCells */


bettipol_t GetBAdress(p)

/*
  This function and the next two are the analogues of the corresponding
  functions for Kazhdan-Lusztig polynomials. They are used to manage
  the tree of betti polynomials.
*/

     bettipol_t p;

{
  register betticell_t *c;
  int m;
  
  c = bettistore;	/* entry point of tree */
  
  while (m = BCompare(p,c->pol))
    {
      if (m > 0) /* p is smaller than c->pol */
	{
	  if (c->left)	/* go to left son */
	    {
	      c = c->left;
	    }
	  else	/* make new cell */
	    {
	      c->left = MakeBSon(p);
	      c = c->left;
	      break;
	    }
	}
      else	/* p is greater than c->pol */
	{
	  if (c->right)	/* go to right son */
	    {
	      c = c->right;
	    }
	  else	/* make new cell */
	    {
	      c->right = MakeBSon(p);
	      c = c->right;
	      break;
	    }
	}
    }
  
  /* if we get here, pol has been found */
  
  return(c->pol);
  
}    /* end of GetBAdress */


BCompare(p,q)

/*
  This is slightly different from the KL case because we may be dealing
  with large unsigned quantities, so we only return a sign.
*/

     bettipol_t p,q;

{
  int i,d;
  
  if (*q - *p)	/* degrees are different */
    {
      if (*p<*q)
	{
	  return(1);
	}
      else
	{
	  return(-1);
	}
    }
  
  d = (int)*p;
  for (i=0;i<=d;i++)
     if (*(++q)-*(++p))
      {
      if (*p<*q)
	{
	  return(1);
	}
      else
	{
	  return(-1);
	}
      }
  
  return(0);
  
}    /* end of BCompare */


betticell_t *MakeBSon(p)

/*

*/

     bettipol_t p;

{
  betticell_t *newcell;
  int i;
  
  CALLOC(newcell,1,betticell_t,;)
    
    if (bettizone->freestore <= *p+1)	/* make room */
      {
	NewBlock(bettizone,sizeof(weylsize_t));
      }
  
  newcell->pol = (bettipol_t)bettizone->firstfree;
  bettizone->firstfree += (*p+2)*sizeof(weylsize_t);
  bettizone->freestore -= *p+2;
  for (i=0;i<= *p+1;i++)
    (newcell->pol)[i] = p[i];
  
  return(newcell);
  
}    /* end of MakeBSon */


BKLPolMult(bp,klq,bpq)

/*
  Multiplies a bettipol_t bp and a klpol_t klq; the result is again
  a bettipol_t. Otherwise identical to BPolMult. Uses the workstring
  bklmult_work.
*/

     bettipol_t bp;
     klpol_t klq;
     bettipol_t bpq;

{
  register weylsize_t k;
  register int i,j;
  
  *bklmult_work = *bp + *klq;  /* degrees add up */

  for (j=0;j<=*bklmult_work;)  /* clean up pq in the appropriate range */
    {
      bklmult_work[++j] = 0;
    }

  for (j=0;j<=*bp;)
    {
      k = bp[++j];
      for (i=0;i<=*klq;i++)
	{
	  bklmult_work[i+j] += k*klq[i+1];
	}
    }

  for (j=0;j<=*bklmult_work+1;j++)
    {
      bpq[j] = bklmult_work[j];
    }
  
}    /* end of BKLPolMult */


DumpFiles()

/*
  Dumps kl_list, klstore and klzone to files, in order to avoid constant
  computation. This is of course aimed at the case of large groups. The
  difficulty is that absolute adresses shoul be transformed into relative
  ones; the ReadFiles function should link absolute adresses back in.
*/

{
  FILE *listf,*storef,*zonef;
  char filename[LINESIZE];
  unsigned long size,count;
  weylnbr_t w;
  rowsize_t rsize,r;
  kldata_t *kl;

  sprintf(filename,"%c%d.store",type,rank);
  storef = fopen(filename,"wb");
  sprintf(filename,"%c%d.zone",type,rank);
  zonef = fopen(filename,"wb");

  size = 0;
  count = 0;

  WriteTree(storef,zonef,klstore,&size,&count);

  fclose(storef);
  fclose(zonef);

  sprintf(filename,"%c%d.list",type,rank);

  listf = fopen(filename,"wb");

  /* listf starts with the sizes of storef and zonef */

  fwrite(&count,sizeof(count),1,listf);
  fwrite(&size,sizeof(size),1,listf);

  for (w=0;w<wsize;w++)
    {
      fwrite(&(kl_list[w].size),sizeof(rowsize_t),1,listf);
    }

  for (w=0;w<wsize;w++)
    {
      if (rsize = kl_list[w].size)
	{
	  kl = (kldata_t *)kl_list[w].row;
	  for (r=0;r<rsize;r++,kl++)  /* make relative adresses */
	    {
	      kl->pol = (klpol_t *)*(kl->pol);
	    }
	  fwrite(kl_list[w].row,sizeof(kldata_t),rsize,listf);
	}
    }

  fclose(listf);

  FreeZone(klzone);

  DestroyList();
  DestroyTree(klstore);

}  /* end of DumpFiles */


WriteTree(storef,zonef,klcell,adress,count)

/*
  Writes klzone recursively in zonef. Avoids having to deal with the block-
  structure of klzone. Transforms the adresses in klstore into relative adresses.
  If count and adress point initially to 0, they will hold respectively
  the number of cells in klstore and the number of klcoeff_t written in zonef.
  The klstore is destroyed along the way.
*/

    FILE *storef,*zonef;
    klcell_t *klcell;
    unsigned long *adress,*count;

{
  klcoeff_t d;
  klcell_t klc;
  
  if (klcell == 0)
    return(0);

  klc = *klcell;

  if (klc.left)  /* write left tree */
    {
      WriteTree(storef,zonef,klc.left,adress,count);

      /* after WriteTree, count points just after the end of the tree;
         the root of the tree is the last element -- but we use count
	 instead of count-1 to distinguish from the case of NULL */
      
      klc.left = (klcell_t *)(*count);
    }

  /* write polynomial */

  fwrite(klc.pol,sizeof(klcoeff_t),klc.pol[0]+2,zonef);

  /* make relative adress */

  d = klc.pol[0]+2;
  klc.pol = (klpol_t)*adress;
  *adress += d;

  if (klc.right)  /* write right tree */
    {
      WriteTree(storef,zonef,klc.right,adress,count);
      klc.right = (klcell_t *)(*count);
    }

  fwrite(&klc,sizeof(klcell_t),1,storef);
  
  klcell->pol = (klpol_t)*count;
  (*count)++;

}  /* end of WriteTree */


ReadFiles()

/*
  Reads in the files Xn.list, Xn.store, Xn.zone
*/

{
  FILE *listf,*storef,*zonef;
  char filename[LINESIZE];
  klpol_t klzone;
  klcell_t *kl0,*klcell;
  unsigned long size,count,j,extrcount;
  weylnbr_t w;
  kldata_t *kl;

  sprintf(filename,"%c%d.list",type,rank);
  listf = fopen(filename,"rb");
  if (listf == NULL)
    {
      return(1);
    }
  sprintf(filename,"%c%d.store",type,rank);
  storef = fopen(filename,"rb");
  if (storef == NULL)
    {
      return(1);
    }
  sprintf(filename,"%c%d.zone",type,rank);
  zonef = fopen(filename,"rb");
  if (zonef == NULL)
    {
      return(1);
    }

  fread(&count,sizeof(count),1,listf);
  fread(&size,sizeof(size),1,listf);

  CALLOC(kl0,count,klcell_t,;)
  CALLOC(klzone,size,klcoeff_t,;)

  fread(kl0,sizeof(klcell_t),count,storef);
  fread(klzone,sizeof(klcoeff_t),size,zonef);

  /* put back absolute adresses in klstore -- recall relative adresses
     are offset by 1 */

  klcell = kl0;

  for (j=0;j<count;j++,klcell++)
    {
      klcell->pol = klzone + (unsigned long)(klcell->pol);
      if (klcell->left)
	{
	  klcell->left = kl0 + (unsigned long)klcell->left-1;
	}
      if (klcell->right)
	{
	  klcell->right = kl0 + (unsigned long)klcell->right-1;
	}
    }

  klstore = klcell-1;  /* root of klstore is the last cell */

  extrcount = 0;

  for (w=0;w<wsize;w++)
    {
      fread(&(kl_list[w].size),sizeof(rowsize_t),1,listf);
      extrcount += kl_list[w].size;
    }

  CALLOC(kl,extrcount,kldata_t,;)
  fread(kl,sizeof(kldata_t),extrcount,listf);

  for (j=0;j<extrcount;j++)  /* put back absolute adresses */
    {
      kl[j].pol = &((kl0 + (unsigned long)kl[j].pol)->pol);
    }

  extrcount = 0;

  for (w=0;w<wsize;w++)
    {
      if (kl_list[w].size)
	{
	  kl_list[w].row = (char *)kl;
	  kl += kl_list[w].size;
	}
      else
	{
	  kl_list[w].row = NULL;
	}
    }

  fclose(listf);
  fclose(storef);
  fclose(zonef);

  return(0);
}    /* end of ReadFiles */


DestroyList()

/*
  Destroys kl_list (for use after dumping to a file). It is assumed that
  kl_list has been constructed through an allocation for each length. Note
  that kl_list itself is not freed.
*/

{
  lengthcount_t p;
  weylnbr_t w;

  w = 0;

  for (p=0;p<=maxlength;p++)
    {
      free(kl_list[w].row);
      w += betti[p];
    }

 }    /* end of DestroyList */


DestroyTree(klcell)

/*
  Destroys the tree klstore. Intended for use after file dump.
*/

     klcell_t *klcell;

{
  if (klcell == NULL)
    return(0);

  DestroyTree(klcell->left);
  DestroyTree(klcell->right);

  free(klcell);
}    /* end of DestroyTree */


klcoeff_t Multiplicity(w,y)

/*
  Gives access to the multiplicities in Verma modules. Of course, this is
  trivial once the KL polynomials have been computed. We have made no 
  effort to be particularly efficient here. The return value is the
  multiplicity of L_y in M_w, where M_w has highest weight w\rho-\rho
*/

	weylnbr_t w,y;

{

	klpol_t *pol;
	klcoeff_t m;

	pol = KLpol(w,y);

	if (pol == NULL)
		{
		return(0);  /* multiplicity is 0 unless w <= y */
		}
	else
		{
		m = Eval1(*pol);
		return(m);
		}

}    /* end of Multiplicity */


static Eval1(p)

/*
  Computes the value of the polynomial p at 1
*/

	klpol_t p;

{

	int val,d,j;

	val = 0;
	d = *p++;

	for (j=0;j<=d;j++)
		{
		val += *p++;
		}

	return(val);

}    /* end of Eval1 */
